# -*- coding: utf-8 -*-
from dataclasses import dataclass, field
import pathlib
import re
from typing import Iterable, Match, Optional

import numpy as np
from PIL import Image

from isic_challenge_scoring.types import ScoreException


@dataclass
class ImagePair:
    truth_file: pathlib.Path
    truth_image: np.ndarray = field(init=False)
    prediction_file: pathlib.Path = field(init=False)
    prediction_image: np.ndarray = field(init=False)
    image_id: str = field(init=False)
    attribute_id: Optional[str] = field(default=None, init=False)

    def parse_image_id(self):
        image_id_match: Match[str] = re.search(r'ISIC_[0-9]{7}', self.truth_file.stem)
        if not image_id_match:
            raise ScoreException(
                f'Internal error: unknown ground truth file: {self.truth_file.name}.'
            )
        self.image_id = image_id_match.group(0)

        attribute_id_match: Match[str] = re.search(r'attribute_([a-z_]+)', self.truth_file.stem)
        if attribute_id_match:
            self.attribute_id = attribute_id_match.group(1)

    def find_prediction_file(self, prediction_path: pathlib.Path):
        image_number: str = self.image_id.split('_')[1]

        if not self.attribute_id:
            prediction_file_candidates = [
                prediction_file
                for prediction_file in prediction_path.iterdir()
                if image_number in prediction_file.stem
            ]
        else:
            prediction_file_candidates = [
                prediction_file
                for prediction_file in prediction_path.iterdir()
                if image_number in prediction_file.stem
                and self.attribute_id in prediction_file.stem
            ]

        if not prediction_file_candidates:
            raise ScoreException(f'No matching submission for: {self.truth_file.name}')
        elif len(prediction_file_candidates) > 1:
            raise ScoreException(f'Multiple matching submissions for: {self.truth_file.name}')

        self.prediction_file = prediction_file_candidates[0]

    def load_truth_image(self):
        self.truth_image = load_segmentation_image(self.truth_file)
        self.truth_image = assert_binary_image(self.truth_image, self.truth_file)

    def load_prediction_image(self):
        self.prediction_image = load_segmentation_image(self.prediction_file)
        if self.prediction_image.shape[0:2] != self.truth_image.shape[0:2]:
            raise ScoreException(
                f'Image {self.prediction_file.name} has dimensions '
                f'{self.prediction_image.shape[0:2]}; expected {self.truth_image.shape[0:2]}.'
            )


def load_segmentation_image(image_path: pathlib.Path) -> np.ndarray:
    """Load a segmentation image as a NumPy array, given a file path."""
    try:
        image = Image.open(str(image_path))
    except Exception as e:
        raise ScoreException(f'Could not decode image "{image_path.name}" because: "{str(e)}"')

    if image.mode == '1':
        # NumPy crashes if a 1-bit (black and white) image is directly
        # coerced to an array
        image = image.convert('L')

    if image.mode != 'L':
        raise ScoreException(f'Image {image_path.name} is not single-channel (greyscale).')

    image = np.array(image)

    return image


def assert_binary_image(image: np.ndarray, image_path: pathlib.Path):
    """Ensure a NumPy array image is binary, correcting if possible."""
    image_values = set(np.unique(image))
    if image_values <= {0, 255}:
        # Expected values
        pass
    elif len(image_values) <= 2:
        # Binary image with high value other than 255 can be corrected
        high_value = (image_values - {0}).pop()
        image /= high_value
        image *= 255
        if set(np.unique(image)) > {0, 255}:
            raise ScoreException(f'Image {image_path.name} contains values other than 0 and 255.')
    else:
        raise ScoreException(f'Image {image_path.name} contains values other than 0 and 255.')

    return image


def iter_image_pairs(
    truth_path: pathlib.Path, prediction_path: pathlib.Path
) -> Iterable[ImagePair]:
    for truth_file in sorted(truth_path.iterdir()):
        if truth_file.name in {'ATTRIBUTION.txt', 'LICENSE.txt'}:
            continue

        image_pair = ImagePair(truth_file=truth_file)
        image_pair.parse_image_id()
        image_pair.find_prediction_file(prediction_path)
        image_pair.load_truth_image()
        image_pair.load_prediction_image()

        yield image_pair



# -*- coding: utf-8 -*-
from typing import TextIO, Tuple

import numpy as np
import pandas as pd

from isic_challenge_scoring.types import ScoreException


def parse_truth_csv(csv_file_stream: TextIO) -> Tuple[pd.DataFrame, pd.DataFrame]:
    table = pd.read_csv(csv_file_stream, header=0)

    table.set_index('image', drop=True, inplace=True, verify_integrity=False)

    # Support legacy truth files
    if 'score_weight' not in table.columns:
        table['score_weight'] = 1.0
    if 'validation_weight' not in table.columns:
        table['validation_weight'] = 1.0
    if 'ISIC_0035068' in table.index:
        # TODO: Move this to ground truth
        table.loc['ISIC_0035068', ['score_weight', 'validation_weight']] = 0.0

    probabilities = table.drop(columns=['score_weight', 'validation_weight'])
    weights = table[['score_weight', 'validation_weight']]

    return probabilities, weights


def parse_csv(csv_file_stream: TextIO, categories: pd.Index) -> pd.DataFrame:
    probabilities = pd.read_csv(csv_file_stream, header=0)

    if 'image' not in probabilities.columns:
        raise ScoreException('Missing column in CSV: "image".')

    probabilities.set_index('image', drop=True, inplace=True, verify_integrity=True)

    missing_columns = categories.difference(probabilities.columns)
    if not missing_columns.empty:
        raise ScoreException(f'Missing columns in CSV: {list(missing_columns)}.')

    extra_columns = probabilities.columns.difference(categories)
    if not extra_columns.empty:
        raise ScoreException(f'Extra columns in CSV: {list(extra_columns)}.')

    # sort by the order in categories
    probabilities = probabilities.reindex(categories, axis='columns')

    missing_rows = probabilities[probabilities.isnull().any(axis='columns')].index
    if not missing_rows.empty:
        raise ScoreException(f'Missing value(s) in CSV for images: {missing_rows.tolist()}.')

    non_float_columns = probabilities.dtypes[
        probabilities.dtypes.apply(lambda x: x != np.float64)
    ].index
    if not non_float_columns.empty:
        raise ScoreException(
            f'CSV contains non-floating-point value(s) in columns: {non_float_columns.tolist()}.'
        )
    # TODO: identify specific failed rows

    out_of_range_rows = probabilities[
        probabilities.applymap(lambda x: x < 0.0 or x > 1.0).any(axis='columns')
    ].index
    if not out_of_range_rows.empty:
        raise ScoreException(
            f'Values in CSV are outside the interval [0.0, 1.0] for images: '
            f'{out_of_range_rows.tolist()}.'
        )

    # TODO: fail on extra columns in data rows

    return probabilities


def validate_rows(truth_probabilities: pd.DataFrame, prediction_probabilities: pd.DataFrame):
    """
    Ensure prediction rows correspond to truth rows.

    Fail when predictionProbabilities is missing rows or has extra rows compared to
    truthProbabilities.
    """
    missing_images = truth_probabilities.index.difference(prediction_probabilities.index)
    if not missing_images.empty:
        raise ScoreException(f'Missing images in CSV: {missing_images.tolist()}.')

    extra_images = prediction_probabilities.index.difference(truth_probabilities.index)
    if not extra_images.empty:
        raise ScoreException(f'Extra images in CSV: {extra_images.tolist()}.')


def sort_rows(probabilities: pd.DataFrame):
    """Sort rows by labels, in-place."""
    probabilities.sort_index(axis='rows', inplace=True)



# -*- coding: utf-8 -*-
import pathlib

import pandas as pd

from isic_challenge_scoring import metrics
from isic_challenge_scoring.confusion import create_binary_confusion_matrix
from isic_challenge_scoring.load_image import iter_image_pairs
from isic_challenge_scoring.types import ScoresType


def score(truth_path: pathlib.Path, prediction_path: pathlib.Path) -> ScoresType:
    confusion_matrics = pd.DataFrame(
        [
            create_binary_confusion_matrix(
                truth_binary_values=image_pair.truth_image > 128,
                prediction_binary_values=image_pair.prediction_image > 128,
                name=(image_pair.image_id, image_pair.attribute_id),
            )
            for image_pair in iter_image_pairs(truth_path, prediction_path)
        ]
    )

    scores = {
        'macro_average': {
            'threshold_jaccard': confusion_matrics.apply(
                metrics.binary_threshold_jaccard, threshold=0.65, axis='columns'
            ).mean(),
            'jaccard': confusion_matrics.apply(metrics.binary_jaccard, axis='columns').mean(),
            'dice': confusion_matrics.apply(metrics.binary_dice, axis='columns').mean(),
            'accuracy': confusion_matrics.apply(metrics.binary_accuracy, axis='columns').mean(),
            'sensitivity': confusion_matrics.apply(
                metrics.binary_sensitivity, axis='columns'
            ).mean(),
            'specificity': confusion_matrics.apply(
                metrics.binary_specificity, axis='columns'
            ).mean(),
        }
    }
    scores['overall'] = scores['macro_average']['threshold_jaccard']

    return scores



# -*- coding: utf-8 -*-
from typing import Dict, List

import numpy as np
import pandas as pd
import sklearn.metrics


def _to_labels(probabilities: pd.DataFrame) -> pd.Series:
    labels = probabilities.idxmax(axis='columns')

    # Find places where there are multiple maximum values
    max_probabilities = probabilities.max(axis='columns')
    is_max: pd.DataFrame = probabilities.eq(max_probabilities, axis='rows')
    number_of_max: pd.Series = is_max.sum(axis='columns')
    multiple_max: pd.Series = number_of_max.gt(1)
    # Set those locations as an 'undecided' label
    labels[multiple_max] = 'undecided'
    # TODO: emit a warning if any are set to 'undecided'

    return labels


def _get_frequencies(labels: pd.Series, weights: pd.Series, categories: pd.Index) -> pd.Series:
    # Directly sum the weights, grouping them by label
    # .reindex sorts this by the order in categories
    return weights.groupby(labels, sort=False).sum().reindex(categories, fill_value=0)


def _label_balanced_multiclass_accuracy(
    truth_labels: pd.Series, prediction_labels: pd.Series, weights: pd.Series, categories: pd.Index
) -> float:
    # See http://scikit-learn.org/dev/modules/model_evaluation.html#balanced-accuracy-score ; in
    # summary, 'sklearn.metrics.balanced_accuracy_score' is for binary classification only, so we
    # need to implement our own; here, we implement a simpler version of "balanced accuracy" than
    # the definitions mentioned by SciKit learn, as it's just a normalization of TP scores by true
    # class proportions
    confusion_matrix = sklearn.metrics.confusion_matrix(
        truth_labels, prediction_labels, labels=categories, sample_weight=weights
    )
    # TODO: try to convert to a DataFrame, for useful debugging labels
    # confusion_matrix = pd.DataFrame(
    #     confusion_matrix,
    #     index=LABELS.map(lambda label: f'true_{label}'),
    #     columns=LABELS.map(lambda label: f'predicted_{label}')
    # )

    true_positive_counts = pd.Series(confusion_matrix.diagonal(), index=categories)

    # These are equal to rows of the confusion matrix
    true_label_frequencies = _get_frequencies(truth_labels, weights, categories)

    balanced_accuracy = true_positive_counts.divide(true_label_frequencies).mean()
    return balanced_accuracy


def balanced_multiclass_accuracy(
    truth_probabilities: pd.DataFrame, prediction_probabilities: pd.DataFrame, weights: pd.Series
) -> float:
    truth_labels = _to_labels(truth_probabilities)
    prediction_labels = _to_labels(prediction_probabilities)
    categories = truth_probabilities.columns

    # This is easier to test
    balanced_accuracy = _label_balanced_multiclass_accuracy(
        truth_labels, prediction_labels, weights, categories
    )
    return balanced_accuracy


def binary_accuracy(cm: pd.Series) -> float:
    return (cm.at['TP'] + cm.at['TN']) / (cm.at['TP'] + cm.at['TN'] + cm.at['FP'] + cm.at['FN'])


def binary_sensitivity(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FN'] == 0:
        # sensitivity can't be calculated if all are negative, so make this metric a freebie
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FN'])


def binary_specificity(cm: pd.Series) -> float:
    if cm.at['TN'] + cm.at['FP'] == 0:
        # specificity can't be calculated if all are positive, so make this metric a freebie
        return 1.0
    else:
        return cm.at['TN'] / (cm.at['TN'] + cm.at['FP'])


def binary_jaccard(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] + cm.at['FN'] == 0:
        # Jaccard is ill-defined if all are negative and the prediction is perfect, but we'll
        # just score that as a perfect answer
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FP'] + cm.at['FN'])


def binary_threshold_jaccard(cm: pd.Series, threshold: float = 0.65) -> float:
    jaccard = binary_jaccard(cm)
    return jaccard if jaccard >= threshold else 0.0


def binary_dice(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] + cm.at['FN'] == 0:
        # Dice is ill-defined if all are negative and the prediction is perfect, but we'll
        # just score that as a perfect answer
        return 1.0
    else:
        return (2 * cm.at['TP']) / ((2 * cm.at['TP']) + cm.at['FP'] + cm.at['FN'])


def binary_ppv(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] == 0:
        # PPV is ill-defined if all predictions are negative; we'll score it as perfect, which
        # doesn't penalize the case where all are truly negative (a good predictor), and is sane
        # for the case where some are truly positive (a limitation of this metric)
        # Note, some other implementations would score the latter case as 0:
        # https://github.com/dice-group/gerbil/wiki/Precision,-Recall-and-F1-measure
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FP'])


def binary_npv(cm: pd.Series) -> float:
    if cm.at['TN'] + cm.at['FN'] == 0:
        # NPV is ill-defined if all predictions are positive; we'll score it as perfect, which
        # doesn't penalize the case where all are truly positive (a good predictor), and is sane
        # for the case where some are truly negative (a limitation of this metric)
        # Note, some other implementations would score the latter case as 0:
        # https://github.com/dice-group/gerbil/wiki/Precision,-Recall-and-F1-measure
        return 1.0
    else:
        return cm.at['TN'] / (cm.at['TN'] + cm.at['FN'])


def auc(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> float:
    auc = sklearn.metrics.roc_auc_score(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )
    return auc


def auc_above_sensitivity(
    truth_probabilities: pd.Series,
    prediction_probabilities: pd.Series,
    weights: pd.Series,
    sensitivity_threshold: float,
) -> float:
    # Get the ROC curve points
    # TODO: We must have both some true and false instances in truthProbabilities
    false_positive_rates, true_positive_rates, thresholds = sklearn.metrics.roc_curve(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )

    # Search for the index along the curve where sensitivity_threshold occurs
    # sensitivity == true_positive_rate
    threshold_index = np.argmax(true_positive_rates >= sensitivity_threshold)

    # Get the corresponding FPR value at the threshold, needed for sklearn
    fpr_threshold = false_positive_rates[threshold_index]
    # Compute the partial AUC over the range [0, fpr_threshold], which is the complement of what we
    # want (i.e. [fpr_threshold, 1]), but this is all sklearn can provide
    if fpr_threshold == 0.0:
        # roc_auc_score's max_fpr requires a value > 0
        complementary_auc = 0.0
    else:
        complementary_auc = sklearn.metrics.roc_auc_score(
            truth_probabilities,
            prediction_probabilities,
            sample_weight=weights,
            max_fpr=fpr_threshold,
        )

    total_auc = sklearn.metrics.roc_auc_score(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )

    # complementary_auc is the left / lower area, and we want the right / upper area
    partial_auc = total_auc - complementary_auc
    return partial_auc


def average_precision(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> float:
    ap = sklearn.metrics.average_precision_score(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )
    return ap


def roc(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> List[Dict[str, float]]:
    fprs, tprs, thresholds = sklearn.metrics.roc_curve(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )
    roc = list(
        map(lambda fpr, tpr, t: {'fpr': fpr, 'tpr': tpr, 'threshold': t}, fprs, tprs, thresholds)
    )
    return roc



# -*- coding: utf-8 -*-
import json
import os
import pathlib
import shutil
import tempfile
from typing import Tuple
import zipfile

from isic_challenge_scoring import task1
from isic_challenge_scoring import task2
from isic_challenge_scoring import task3
from isic_challenge_scoring.types import ScoreException, ScoresType


def extract_zip(zip_path: pathlib.Path, output_path: pathlib.Path, flatten: bool = True):
    """Extract a zip file, optionally flattening it into a single directory."""
    try:
        with zipfile.ZipFile(zip_path) as zf:
            if flatten:
                for member_info in zf.infolist():
                    member_name = member_info.filename
                    if member_name.startswith('__MACOSX'):
                        # Ignore Mac OS X metadata
                        continue

                    member_base_name = os.path.basename(member_name)
                    if not member_base_name:
                        # Skip directories
                        continue

                    member_output_path = output_path / member_base_name

                    with zf.open(member_info) as input_stream, member_output_path.open(
                        'wb'
                    ) as output_stream:
                        shutil.copyfileobj(input_stream, output_stream)
            else:
                zf.extractall(output_path)
    except zipfile.BadZipfile as e:
        raise ScoreException(f'Could not read ZIP file "{zip_path.name}": {str(e)}.')


def unzip_all(
    input_path: pathlib.Path, allow_manuscript_directory: bool = False
) -> Tuple[pathlib.Path, tempfile.TemporaryDirectory]:
    """
    Extract / copy all files in directory.

    Validates that the path contains exactly one file. Optionally allow an 'Abstract' directory to
    exist which contains exactly one manuscript file. Return a path to the extracted content.
    """
    input_files = [f for f in input_path.iterdir() if f.is_file()]
    input_dirs = [f for f in input_path.iterdir() if f.is_dir()]

    if len(input_files) > 1:
        raise ScoreException('Multiple files submitted. Exactly one ZIP file should be submitted.')
    elif len(input_files) < 1:
        raise ScoreException('No files submitted. Exactly one ZIP file should be submitted.')

    input_file = input_files[0]

    manuscript_file = None

    if allow_manuscript_directory:
        if len(input_dirs) > 1:
            raise ScoreException('Internal error: multiple directories found.')
        elif len(input_dirs) == 1:
            input_dir = input_dirs[0]
            if input_dir.name != 'Abstract':
                raise ScoreException(
                    f'Internal error: unexpected directory found: {input_dir.name}.'
                )

            manuscript_files = list(input_dir.iterdir())
            if not manuscript_files:
                raise ScoreException('Empty manuscript directory found.')
            elif len(manuscript_files) > 1:
                raise ScoreException('Multiple files found in manuscript directory.')

            manuscript_file = manuscript_files[0]
    elif input_dirs:
        # Expect only files
        raise ScoreException('Internal error: unexpected directory found.')

    output_temp_dir = tempfile.TemporaryDirectory()
    output_path = pathlib.Path(output_temp_dir.name)

    if input_file.suffix.lower() == '.zip':
        extract_zip(input_file, output_path)
    else:
        shutil.copy(input_file, output_path)

    if manuscript_file is not None:
        shutil.copy(manuscript_file, output_path)

    return output_path, output_temp_dir


def ensure_manuscript(prediction_path: pathlib.Path):
    manuscript_file_count = sum(
        manuscript_file.suffix.lower() == '.pdf' for manuscript_file in prediction_path.iterdir()
    )
    if manuscript_file_count > 1:
        raise ScoreException(
            'Multiple PDFs submitted. Exactly one PDF file, containing the descriptive manuscript, '
            'must included in the submission.'
        )
    elif manuscript_file_count < 1:
        raise ScoreException(
            'No PDF submitted. Exactly one PDF file, containing the descriptive manuscript, '
            'must included in the submission.'
        )


def score_all(
    truth_input_path: pathlib.Path,
    prediction_input_path: pathlib.Path,
    task_num: int,
    require_manuscript: bool,
):
    # Unzip zip files contained in the input folders
    truth_path, truth_temp_dir = unzip_all(truth_input_path)

    prediction_path, prediction_temp_dir = unzip_all(
        prediction_input_path, allow_manuscript_directory=True
    )

    if require_manuscript:
        ensure_manuscript(prediction_path)

    if task_num == 1:
        score = task1.score
    elif task_num == 2:
        score = task2.score
    elif task_num == 3:
        score = task3.score
    else:
        raise ScoreException(f'Internal error: unknown ground truth phase number: {task_num}.')
    scores: ScoresType = score(truth_path, prediction_path)

    # Output in Covalic format
    print(
        json.dumps(
            [
                {
                    'dataset': dataset,
                    'metrics': [
                        {'name': metric_name, 'value': metric_value}
                        for metric_name, metric_value in metrics.items()
                    ],
                }
                for dataset, metrics in scores.items()
            ]
        )
    )

    truth_temp_dir.cleanup()
    prediction_temp_dir.cleanup()



# -*- coding: utf-8 -*-
import pathlib
import sys

import click
import click_pathlib

from isic_challenge_scoring import score_all
from isic_challenge_scoring.types import ScoreException


DirectoryPath = click_pathlib.Path(exists=True, file_okay=False, dir_okay=True, readable=True)


@click.command(name='isic-challenge-scoring', help='ISIC Challenge submission scoring')
@click.option(
    'truth_input_path',
    '--groundtruth',
    required=True,
    type=DirectoryPath,
    help='path to the ground truth directory',
)
@click.option(
    'prediction_input_path',
    '--submission',
    required=True,
    type=DirectoryPath,
    help='path to the submission directory',
)
@click.option(
    'task_num',
    '--task',
    required=True,
    type=click.IntRange(1, 3),
    help='challenge task number (1, 2, or 3)',
)
@click.option('--require-manuscript', 'require_manuscript', is_flag=True, default=False)
def main(
    truth_input_path: pathlib.Path,
    prediction_input_path: pathlib.Path,
    task_num: int,
    require_manuscript: bool,
):
    try:
        score_all(truth_input_path, prediction_input_path, task_num, require_manuscript)
    except ScoreException as e:
        covalic_error_prefix = 'covalic.error: '
        print(covalic_error_prefix + str(e), file=sys.stderr)
        sys.exit(1)


main()



# -*- coding: utf-8 -*-
import pathlib
import re
from typing import KeysView

import numpy as np
import pandas as pd

from isic_challenge_scoring import metrics
from isic_challenge_scoring.confusion import create_binary_confusion_matrix
from isic_challenge_scoring.load_csv import parse_csv, parse_truth_csv, sort_rows, validate_rows
from isic_challenge_scoring.types import ScoreException, ScoresType


def compute_metrics(truth_file_stream, prediction_file_stream) -> ScoresType:
    truth_probabilities, truth_weights = parse_truth_csv(truth_file_stream)
    categories = truth_probabilities.columns
    prediction_probabilities = parse_csv(prediction_file_stream, categories)

    validate_rows(truth_probabilities, prediction_probabilities)

    sort_rows(truth_probabilities)
    sort_rows(prediction_probabilities)

    scores: ScoresType = {}
    for category in categories:
        truth_category_probabilities: pd.Series = truth_probabilities[category]
        prediction_category_probabilities: pd.Series = prediction_probabilities[category]

        truth_binary_values: pd.Series = truth_category_probabilities.gt(0.5)
        prediction_binary_values: pd.Series = prediction_category_probabilities.gt(0.5)

        category_cm = create_binary_confusion_matrix(
            truth_binary_values=truth_binary_values.to_numpy(),
            prediction_binary_values=prediction_binary_values.to_numpy(),
            weights=truth_weights.score_weight.to_numpy(),
            name=category,
        )

        scores[category] = {
            'accuracy': metrics.binary_accuracy(category_cm),
            'sensitivity': metrics.binary_sensitivity(category_cm),
            'specificity': metrics.binary_specificity(category_cm),
            'dice': metrics.binary_dice(category_cm),
            'ppv': metrics.binary_ppv(category_cm),
            'npv': metrics.binary_npv(category_cm),
            'auc': metrics.auc(
                truth_category_probabilities,
                prediction_category_probabilities,
                truth_weights.score_weight,
            ),
            'auc_sens_80': metrics.auc_above_sensitivity(
                truth_category_probabilities,
                prediction_category_probabilities,
                truth_weights.score_weight,
                0.80,
            ),
            'ap': metrics.average_precision(
                truth_category_probabilities,
                prediction_category_probabilities,
                truth_weights.score_weight,
            ),
            'roc': metrics.roc(
                truth_category_probabilities,
                prediction_category_probabilities,
                truth_weights.score_weight,
            ),
        }

    # Compute averages for all per-category metrics
    per_category_metrics: KeysView[str] = next(iter(scores.values())).keys()
    scores['macro_average'] = {
        metric: float(np.mean([scores[category][metric] for category in categories]))
        for metric in per_category_metrics
        if metric != 'roc'
    }

    # Compute multi-category aggregate metrics
    scores['aggregate'] = {
        'balanced_accuracy': metrics.balanced_multiclass_accuracy(
            truth_probabilities, prediction_probabilities, truth_weights.score_weight
        )
    }

    scores['overall'] = scores['aggregate']['balanced_accuracy']
    scores['validation'] = metrics.balanced_multiclass_accuracy(
        truth_probabilities, prediction_probabilities, truth_weights.validation_weight
    )

    return scores


def score(truth_path: pathlib.Path, prediction_path: pathlib.Path) -> ScoresType:
    for truth_file in truth_path.iterdir():
        if re.match(r'^ISIC.*GroundTruth\.csv$', truth_file.name):
            break
    else:
        raise ScoreException('Internal error, truth file could not be found.')

    prediction_files = [
        prediction_file
        for prediction_file in prediction_path.iterdir()
        if prediction_file.suffix.lower() == '.csv'
    ]
    if len(prediction_files) > 1:
        raise ScoreException(
            'Multiple prediction files submitted. Exactly one CSV file should be submitted.'
        )
    elif len(prediction_files) < 1:
        raise ScoreException(
            'No prediction files submitted. Exactly one CSV file should be submitted.'
        )
    prediction_file = prediction_files[0]

    with truth_file.open('rb') as truth_file_stream, prediction_file.open(
        'rb'
    ) as prediction_file_stream:
        return compute_metrics(truth_file_stream, prediction_file_stream)



# -*- coding: utf-8 -*-
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd


def create_binary_confusion_matrix(
    truth_binary_values: np.ndarray,
    prediction_binary_values: np.ndarray,
    weights: Optional[np.ndarray] = None,
    name: Optional[Union[str, Tuple[str, ...]]] = None,
) -> pd.Series:
    # This implementation is:
    # ~30x faster than sklearn.metrics.confusion_matrix
    # ~25x faster than sklearn.metrics.confusion_matrix(labels=[False, True])
    # ~6x faster than pandas.crosstab
    truth_binary_values = truth_binary_values.ravel()
    prediction_binary_values = prediction_binary_values.ravel()
    if weights is not None:
        weights = weights.ravel()

    truth_binary_negative_values = 1 - truth_binary_values
    test_binary_negative_values = 1 - prediction_binary_values

    true_positives = np.logical_and(truth_binary_values, prediction_binary_values)
    true_negatives = np.logical_and(truth_binary_negative_values, test_binary_negative_values)
    false_positives = np.logical_and(truth_binary_negative_values, prediction_binary_values)
    false_negatives = np.logical_and(truth_binary_values, test_binary_negative_values)

    if weights is not None:
        true_positives = true_positives * weights
        true_negatives = true_negatives * weights
        false_positives = false_positives * weights
        false_negatives = false_negatives * weights

    true_positive = np.sum(true_positives)
    true_negative = np.sum(true_negatives)
    false_positive = np.sum(false_positives)
    false_negative = np.sum(false_negatives)

    # Storing the matrix as a Series instead of a DataFrame makes it easier to reference elements
    # and aggregate multiple matrices
    cm = pd.Series(
        {'TP': true_positive, 'TN': true_negative, 'FP': false_positive, 'FN': false_negative},
        name=name,
    )

    return cm


def normalize_confusion_matrix(cm: pd.Series) -> pd.Series:
    return cm / cm.sum()



import os
import pathlib

from setuptools import find_packages, setup


def prerelease_local_scheme(version):
    """
    Return local scheme version unless building on master in CircleCI.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    gitlab_branch = os.getenv('CI_COMMIT_REF_NAME')
    circleci_branch = os.getenv('CIRCLE_BRANCH')
    if (gitlab_branch or circleci_branch) == 'master':
        return ''
    else:
        return get_local_node_and_date(version)


with (pathlib.Path(__file__).parent / 'README.md').open() as description_stream:
    long_description = description_stream.read()


setup(
    name='isic_challenge_scoring',
    description='Submission scoring for the ISIC Challenge',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ImageMarkup/isic-challenge-scoring',
    license='Apache 2.0',
    maintainer='ISIC Archive',
    maintainer_email='admin@isic-archive.com',
    packages=find_packages(exclude=['tests']),
    python_requires='>=3.7.0',
    install_requires=[
        'click',
        'click-pathlib',
        'numpy',
        'pandas',
        'pillow',
        'scipy',
        'scikit-learn',
    ],
    use_scm_version={'local_scheme': prerelease_local_scheme},
)
